@@ -12,17 +12,13 @@
 
 """
 
-from __future__ import with_statement
-
 __all__ = ['Freezer', 'walk_directory', 'relative_url_for']
 
-VERSION = '0.10'
-
+VERSION = '0.17'
 
+import datetime
 import os.path
 import mimetypes
-import urlparse
-import urllib
 import warnings
 import collections
 import posixpath
@@ -30,46 +26,54 @@ from fnmatch import fnmatch
 from unicodedata import normalize
 from threading import Lock
 from contextlib import contextmanager
+from posixpath import relpath as posix_relpath
+
+from collections import namedtuple
+try:
+    from collections.abc import Mapping  # Python 3
+except ImportError:
+    from collections import Mapping  # Python 2.7
+
+try:
+    from urlparse import urlsplit
+    from werkzeug.urls import url_unquote as unquote
+except ImportError:  # Python 3
+    from urllib.parse import urlsplit, unquote
 
 from werkzeug.exceptions import HTTPException
 from flask import (Flask, Blueprint, url_for, request, send_from_directory,
                    redirect)
 
 try:
-    from collections import Mapping
-    def is_mapping(obj):
-        return isinstance(obj, Mapping)
-except ImportError:
-    # Python 2.5, no Abstract Base Classes. Default to duck-typing.
-    def is_mapping(obj):
-        return hasattr(obj, 'keys')
-
-try:
-    from posixpath import relpath as posix_relpath
-except ImportError:
-    # Python 2.5
-    def posix_relpath(path, start):
-        sep = posixpath.sep
-        start_list = [x for x in posixpath.abspath(start).split(sep) if x]
-        path_list = [x for x in posixpath.abspath(path).split(sep) if x]
-
-        # Work out how much of the filepath is shared by start and path.
-        i = len(posixpath.commonprefix([start_list, path_list]))
-
-        rel_list = [posixpath.pardir] * (len(start_list)-i) + path_list[i:]
-        if not rel_list:
-            return posixpath.curdir
-        return posixpath.join(*rel_list)
+    unicode
+except NameError:  # Python 3
+    unicode = str
+    basestring = str
 
 
-class MissingURLGeneratorWarning(Warning):
+class FrozenFlaskWarning(Warning):
     pass
 
 
-class MimetypeMismatchWarning(Warning):
+class MissingURLGeneratorWarning(FrozenFlaskWarning):
     pass
 
 
+class MimetypeMismatchWarning(FrozenFlaskWarning):
+    pass
+
+
+class NotFoundWarning(FrozenFlaskWarning):
+    pass
+
+
+class RedirectWarning(FrozenFlaskWarning):
+    pass
+
+
+Page = namedtuple('Page', 'url path')
+
+
 class Freezer(object):
     """
     :param app: your application or None if you use :meth:`init_app`
@@ -110,15 +114,20 @@ class Freezer(object):
             self.url_for_logger = UrlForLogger(app)
             app.config.setdefault('FREEZER_DESTINATION', 'build')
             app.config.setdefault('FREEZER_DESTINATION_IGNORE', [])
-            app.config.setdefault('FREEZER_BASE_URL', 'http://localhost/')
+            app.config.setdefault('FREEZER_STATIC_IGNORE', [])
+            app.config.setdefault('FREEZER_BASE_URL', None)
             app.config.setdefault('FREEZER_REMOVE_EXTRA_FILES', True)
             app.config.setdefault('FREEZER_DEFAULT_MIMETYPE',
                                   'application/octet-stream')
             app.config.setdefault('FREEZER_IGNORE_MIMETYPE_WARNINGS', False)
             app.config.setdefault('FREEZER_RELATIVE_URLS', False)
+            app.config.setdefault('FREEZER_IGNORE_404_NOT_FOUND', False)
+            app.config.setdefault('FREEZER_REDIRECT_POLICY', 'follow')
+            app.config.setdefault('FREEZER_SKIP_EXISTING', False)
 
     def register_generator(self, function):
-        """Register a function as an URL generator.
+        """
+        Register a function as an URL generator.
 
         The function should return an iterable of URL paths or
         ``(endpoint, values)`` tuples to be used as
@@ -142,8 +151,22 @@ class Freezer(object):
             unicode(self.app.config['FREEZER_DESTINATION'])
         )
 
-    def freeze(self):
-        """Clean the destination and build all URLs from generators."""
+    def freeze_yield(self):
+        """Like :meth:`freeze`, but yields information about pages as they are
+        being processed. Yields :func:`namedtuples <collections.namedtuple>`
+        ``(url, path)``. This can be used to display progress information,
+        such as printing the information to standard output, or even more
+        sophisticated, e.g. with a :func:`progressbar <click.progressbar>`::
+
+            import click
+
+            with click.progressbar(
+                    freezer.freeze_yield(),
+                    item_show_func=lambda p: p.url if p else 'Done!') as urls:
+                for url in urls:
+                    # everything is already happening, just pass
+                    pass
+        """
         remove_extra = self.app.config['FREEZER_REMOVE_EXTRA_FILES']
         if not os.path.isdir(self.root):
             os.makedirs(self.root)
@@ -157,14 +180,15 @@ class Freezer(object):
         seen_endpoints = set()
         built_files = set()
 
-        for url, endpoint in self._generate_all_urls():
+        for url, endpoint, last_modified in self._generate_all_urls():
             seen_endpoints.add(endpoint)
             if url in seen_urls:
                 # Don't build the same URL more than once
                 continue
             seen_urls.add(url)
-            new_filename = self._build_one(url)
+            new_filename = self._build_one(url, last_modified)
             built_files.add(normalize('NFC', new_filename))
+            yield Page(url, os.path.relpath(new_filename, self.root))
 
         self._check_endpoints(seen_endpoints)
         if remove_extra:
@@ -175,7 +199,10 @@ class Freezer(object):
                 if not os.listdir(parent):
                     # The directory is now empty, remove it.
                     os.removedirs(parent)
-        return seen_urls
+
+    def freeze(self):
+        """Clean the destination and build all URLs from generators."""
+        return set(page.url for page in self.freeze_yield())
 
     def all_urls(self):
         """
@@ -187,7 +214,7 @@ class Freezer(object):
             generated from :func:`~flask.url_for` calls will not be included
             here.
         """
-        for url, _endpoint in self._generate_all_urls():
+        for url, _endpoint, last_modified in self._generate_all_urls():
             yield url
 
     def _script_name(self):
@@ -195,32 +222,38 @@ class Freezer(object):
         Return the path part of FREEZER_BASE_URL, without trailing slash.
         """
         base_url = self.app.config['FREEZER_BASE_URL']
-        return urlparse.urlsplit(base_url).path.rstrip('/')
+        return urlsplit(base_url or '').path.rstrip('/')
 
     def _generate_all_urls(self):
         """
-        Run all generators and yield (url, enpoint) tuples.
+        Run all generators and yield (url, endpoint) tuples.
         """
         script_name = self._script_name()
         url_encoding = self.app.url_map.charset
         url_generators = list(self.url_generators)
         url_generators += [self.url_for_logger.iter_calls]
         # A request context is required to use url_for
-        with self.app.test_request_context(base_url=script_name):
+        with self.app.test_request_context(base_url=script_name or None):
             for generator in url_generators:
                 for generated in generator():
                     if isinstance(generated, basestring):
                         url = generated
                         endpoint = None
+                        last_modified = None
                     else:
-                        if is_mapping(generated):
+                        if isinstance(generated, Mapping):
                             values = generated
                             # The endpoint defaults to the name of the
                             # generator function, just like with Flask views.
                             endpoint = generator.__name__
+                            last_modified = None
                         else:
                             # Assume a tuple.
-                            endpoint, values = generated
+                            if len(generated)==2:
+                                endpoint, values = generated
+                                last_modified = None
+                            else:
+                                endpoint, values, last_modified = generated
                         url = url_for(endpoint, **values)
                         assert url.startswith(script_name), (
                             'url_for returned an URL %r not starting with '
@@ -229,8 +262,8 @@ class Freezer(object):
                         )
                         url = url[len(script_name):]
                     # flask.url_for "quotes" URLs, eg. a space becomes %20
-                    url = urllib.unquote(url)
-                    parsed_url = urlparse.urlsplit(url)
+                    url = unquote(url)
+                    parsed_url = urlsplit(url)
                     if parsed_url.scheme or parsed_url.netloc:
                         raise ValueError('External URLs not supported: ' + url)
 
@@ -238,11 +271,11 @@ class Freezer(object):
                     url = parsed_url.path
                     if not isinstance(url, unicode):
                         url = url.decode(url_encoding)
-                    yield url, endpoint
+                    yield url, endpoint, last_modified
 
     def _check_endpoints(self, seen_endpoints):
         """
-        Warn if some of the app's enpoints are not in seen_endpoints.
+        Warn if some of the app's endpoints are not in seen_endpoints.
         """
         get_endpoints = set(
             rule.endpoint for rule in self.app.url_map.iter_rules()
@@ -255,32 +288,54 @@ class Freezer(object):
 
         if not_generated_endpoints:
             warnings.warn(
-                'Nothing frozen for endpoints %s. Did you forget an URL '
+                'Nothing frozen for endpoints %s. Did you forget a URL '
                 'generator?' % ', '.join(
                     unicode(e) for e in not_generated_endpoints),
                 MissingURLGeneratorWarning,
                 stacklevel=3)
 
-    def _build_one(self, url):
-        """Get the given ``url`` from the app and write the matching file.
-        """
+    def _build_one(self, url, last_modified=None):
+        """Get the given ``url`` from the app and write the matching file."""
         client = self.app.test_client()
         base_url = self.app.config['FREEZER_BASE_URL']
+        redirect_policy = self.app.config['FREEZER_REDIRECT_POLICY']
+        follow_redirects = redirect_policy == 'follow'
+        ignore_redirect = redirect_policy == 'ignore'
+
+        destination_path = self.urlpath_to_filepath(url)
+        filename = os.path.join(self.root, *destination_path.split('/'))
+
+        skip = self.app.config['FREEZER_SKIP_EXISTING']
+        if callable(skip):
+            skip = skip(url, filename)
+        if os.path.isfile(filename):
+            mtime = datetime.datetime.fromtimestamp(os.path.getmtime(filename))
+            if (last_modified is not None and mtime>=last_modified) or skip:
+                return filename
 
         with conditional_context(self.url_for_logger, self.log_url_for):
             with conditional_context(patch_url_for(self.app),
                                      self.app.config['FREEZER_RELATIVE_URLS']):
-                response = client.get(url, follow_redirects=True,
+                response = client.get(url, follow_redirects=follow_redirects,
                                       base_url=base_url)
 
         # The client follows redirects by itself
         # Any other status code is probably an error
-        if not(response.status_code == 200):
-            raise ValueError('Unexpected status %r on URL %s' \
-                % (response.status, url))
-
-        destination_path = self.urlpath_to_filepath(url)
-        filename = os.path.join(self.root, *destination_path.split('/'))
+        # except we explicitly want 404 errors to be skipped
+        # (eg. while application is in development)
+        ignore_404 = self.app.config['FREEZER_IGNORE_404_NOT_FOUND']
+        if response.status_code != 200:
+            if response.status_code == 404 and ignore_404:
+                warnings.warn('Ignored %r on URL %s' % (response.status, url),
+                              NotFoundWarning,
+                              stacklevel=3)
+            elif response.status_code in (301, 302) and ignore_redirect:
+                warnings.warn('Ignored %r on URL %s' % (response.status, url),
+                              RedirectWarning,
+                              stacklevel=3)
+            else:
+                raise ValueError('Unexpected status %r on URL %s' \
+                    % (response.status, url))
 
         if not self.app.config['FREEZER_IGNORE_MIMETYPE_WARNINGS']:
             # Most web servers guess the mime type of static files by their
@@ -317,6 +372,7 @@ class Freezer(object):
             with open(filename, 'wb') as fd:
                 fd.write(content)
 
+        response.close()
         return filename
 
     def urlpath_to_filepath(self, path):
@@ -330,7 +386,8 @@ class Freezer(object):
         return path[1:]
 
     def serve(self, **options):
-        """Run an HTTP server on the result of the build.
+        """
+        Run an HTTP server on the result of the build.
 
         :param options: passed to :meth:`flask.Flask.run`.
         """
@@ -390,10 +447,11 @@ class Freezer(object):
             view = self.app.view_functions[endpoint]
             app_or_blueprint = method_self(view)
             root = app_or_blueprint.static_folder
+            ignore = self.app.config['FREEZER_STATIC_IGNORE']
             if root is None or not os.path.isdir(root):
                 # No 'static' directory for this app/blueprint.
                 continue
-            for filename in walk_directory(root):
+            for filename in walk_directory(root, ignore=ignore):
                 yield endpoint, {'filename': filename}
 
     def no_argument_rules_urls(self):
@@ -413,7 +471,7 @@ def walk_directory(root, ignore=()):
     :param ignore:
         A list of :mod:`fnmatch` patterns.
         As in ``.gitignore`` files,
-        patterns that contains a slash are matched against the whole path,
+        patterns that contain a slash are matched against the whole path,
         others against individual slash-separated parts.
 
     """
@@ -438,10 +496,10 @@ def walk_directory(root, ignore=()):
 
 @contextmanager
 def patch_url_for(app):
-    """Patche ``url_for`` in Jinja globals to use :func:`relative_url_for`.
-
-    This is a context manager, to use in a ``with`` statement.
+    """
+    Patches ``url_for`` in Jinja globals to use :func:`relative_url_for`.
 
+    This is a context manager, to be used in a ``with`` statement.
     """
     previous_url_for = app.jinja_env.globals['url_for']
     app.jinja_env.globals['url_for'] = relative_url_for
@@ -462,13 +520,12 @@ def relative_url_for(endpoint, **values):
 
     URLs that would otherwise end with ``/`` get ``index.html`` appended,
     as Frozen-Flask does in filenames. Because of this behavior, this function
-    should only with Frozen-Flask, not when running the application in
+    should only be used with Frozen-Flask, not when running the application in
     :meth:`app.run() <flask.Flask.run>` or another WSGI sever.
 
     If the ``FREEZER_RELATIVE_URLS`` `configuration`_ is True, Frozen-Flask
     will automatically patch the application's Jinja environment so that
     ``url_for`` in templates is this function.
-
     """
     url = url_for(endpoint, **values)
 
@@ -476,8 +533,11 @@ def relative_url_for(endpoint, **values):
     if not url.startswith('/'):
         return url
 
+    url, fragment_sep, fragment = url.partition('#')
+    url, query_sep, query = url.partition('?')
     if url.endswith('/'):
         url += 'index.html'
+    url += query_sep + query + fragment_sep + fragment
 
     request_path = request.path
     if not request_path.endswith('/'):
